@@ -25,12 +25,16 @@ import org.apache.flink.util.TestLogger;
 
 import org.apache.kafka.clients.producer.ProducerRecord;
 import org.apache.kafka.common.Configurable;
+import org.apache.kafka.common.header.Header;
+import org.apache.kafka.common.header.internals.RecordHeader;
+import org.apache.kafka.common.header.internals.RecordHeaders;
 import org.apache.kafka.common.serialization.Deserializer;
 import org.apache.kafka.common.serialization.StringDeserializer;
 import org.apache.kafka.common.serialization.StringSerializer;
 import org.junit.Before;
 import org.junit.Test;
 
+import java.nio.charset.StandardCharsets;
 import java.util.Arrays;
 import java.util.Collections;
 import java.util.HashMap;
@@ -143,6 +147,27 @@ public class KafkaRecordSerializationSchemaBuilderTest extends TestLogger {
         assertThat(opened.get()).isTrue();
     }
 
+    @Test
+    public void testSerializeRecordWithHeaderProvider() throws Exception {
+        final HeaderProvider<String> headerProvider =
+                (ignored) ->
+                        new RecordHeaders(
+                                Collections.singletonList(new RecordHeader("a", "a".getBytes())));
+
+        final KafkaRecordSerializationSchema<String> schema =
+                KafkaRecordSerializationSchema.builder()
+                        .setTopic(DEFAULT_TOPIC)
+                        .setValueSerializationSchema(new SimpleStringSchema())
+                        .setHeaderProvider(headerProvider)
+                        .build();
+        final ProducerRecord<byte[], byte[]> record = schema.serialize("a", null, null);
+        assertThat(record).isNotNull();
+        assertThat(record.headers())
+                .singleElement()
+                .extracting(Header::key, Header::value)
+                .containsExactly("a", "a".getBytes(StandardCharsets.UTF_8));
+    }
+
     @Test
     public void testSerializeRecordWithKey() {
         final SerializationSchema<String> serializationSchema = new SimpleStringSchema();
